import collections
import heapq
import json

from django.core.management.base import CommandError
from django.core.management.commands.loaddata import Command as FixtureFinder

from .rudiments import identity


class FixtureAccumulator:
    """
    Keeps record of fixtures in order of priority.

    >>> p = FixtureAccumulator()
    >>> p.add("harrie", "harrie/joekie")
    >>> p.add("catalogue", "catalogue/i18nproduct", 2)
    >>> p.add("harrie", "harrie/jantje", 7)
    >>> p.add("catalogue", "catalogue/product")
    >>> p.add("harrie", "harrie/jakkie", 1)
    >>> p
    <DefaultFixtureDict {
      "harrie": [
        [
          0,
          "harrie/joekie"
        ],
        [
          7,
          "harrie/jantje"
        ],
        [
          1,
          "harrie/jakkie"
        ]
      ],
      "catalogue": [
        [
          0,
          "catalogue/product"
        ],
        [
          2,
          "catalogue/i18nproduct"
        ]
      ]
    }>
    >>> list(p)
    ['harrie/jantje', 'catalogue/i18nproduct']
    >>> p.keys()
    dict_keys(['harrie', 'catalogue'])
    >>> list(p.items())
    [('harrie', 'harrie/jantje'), ('catalogue', 'catalogue/i18nproduct')]
    """

    def __init__(self):
        self.storage = collections.defaultdict(list)

    def __repr__(self):
        return "<DefaultFixtureDict %s>" % json.dumps(self.storage, indent=2)

    def add(self, key, fixture_path, priority=0):
        heapq.heappush(self.storage[key], (priority, fixture_path))

    def keys(self):
        return self.storage.keys()

    def items(self):
        for key, fixtures in self.storage.items():
            (fixture,) = heapq.nlargest(1, fixtures)
            yield (key, fixture[1])

    def __iter__(self):
        for _, fixtures in self.storage.items():
            (fixture,) = heapq.nlargest(1, fixtures)
            yield fixture[1]


def find_fixture(fixture_name):
    """
    Finds a fixture and returns the path.
    Also works with strange fixtures that can not be loaded at all like
    "original_images.tar.gz" or "books.essential.csv"
    """

    class Liar(collections.defaultdict):
        def __contains__(self, name):
            return True

    fixture_finder = FixtureFinder()
    fixture_finder.verbosity = 0
    fixture_finder.using = None
    fixture_finder.app_label = None
    fixture_finder.serialization_formats = Liar(lambda: True)
    fixture_finder.compression_formats = Liar(lambda: identity)

    try:
        for path, _, _ in fixture_finder.find_fixtures(fixture_name):
            return path
    except CommandError:
        return None



from functools import partial


def placeholderify(form=None, fields=None):
    """
    A decorator for Django forms that sets a `placeholder` attribute to all
    fields. Each field's label is used as a placeholder.

    Use it like so:

    @placeholderify
    class MyForm(forms.Form):
        name = forms.CharField(label='Your name')
        email = forms.EmailField(label='Your email')
    Doing that, the `name` field will render like so:
        <input ... placeholder="Your name">
    """
    if form is None:
        return partial(placeholderify, fields=fields)

    class WrappedForm(form):
        def __init__(self, *args, **kwargs):
            super(WrappedForm, self).__init__(*args, **kwargs)

            if fields is None:
                override = self.fields.values()
            else:
                override = [
                    field for name, field in self.fields.items() if name in fields
                ]

            for field in override:
                field.widget.attrs.setdefault("placeholder", field.label)

    return WrappedForm



import copy
import importlib
import logging
import threading
from importlib import import_module
import sys

from configtype.jsonconfig import JsonFinder
from django.utils.functional import cached_property

logger = logging.getLogger(__name__)


def _get_jsonfinder():
    for finder in sys.meta_path:
        if isinstance(finder, JsonFinder):
            return finder
    return None


def _load_config_values():
    config_module_name = f"{__package__}.json"
    try:
        config_module = importlib.import_module(config_module_name)
    except ImportError:
        if _get_jsonfinder() is None:
            logger.debug(
                "No configtype search paths configured for %s; using empty defaults",
                config_module_name,
            )
        else:
            logger.debug(
                "No JSON config found for %s via configured search paths; using empty defaults",
                config_module_name,
            )
        return {}

    values = {}
    for key in dir(config_module):
        if key.startswith("__"):
            continue
        values[key] = copy.deepcopy(getattr(config_module, key))
    return values


class FenderConfig(threading.local):
    """
    Interface for accessing the fender settings from the json config
    """

    settings = {}
    ocyan_plugins = []

    def sections(self):
        return self.settings.keys()

    def has_section(self, name):
        return name in self.settings

    def has_option(self, section, option):
        return self.has_section(section) and option in self.settings[section]

    def get_section(self, section):
        return self.settings.get(section, None)

    def get(self, section, option, fallback=None):
        if self.has_option(section, option):
            return self.settings[section][option]

        return fallback

    @cached_property
    def is_webshop(self):
        # if it doesn't have the barebone plugin, it is a webshop
        # pylint: disable=no-member
        return not any(["barebone" in x for x in self.ocyan_plugins])

    @cached_property
    def i18n_enabled(self):
        try:
            import_module("ocyan.i18n")
        except ImportError:
            return False

        return "i18n" in self.settings


for _key, _value in _load_config_values().items():
    setattr(FenderConfig, _key, _value)


config = FenderConfig()



try:
    from django.contrib.staticfiles.templatetags.staticfiles import (
        static as django_static,
    )
except ImportError:
    from django.templatetags.static import static as django_static


def static(path):
    """
    Use the Django builtin static file resolver to return an absolute path
    usable as CSS url() argument. Sass equivalent of the 'static' template
    tag.
    """
    return '"{}"'.format(django_static(path))



import doctest
import importlib
import json
import sys
import tempfile
import unittest
from pathlib import Path

import ocyan.core.utils.appextension
import ocyan.core.utils.cache
import ocyan.core.utils.csp
import ocyan.core.utils.currency
import ocyan.core.utils.dirtyhacks
import ocyan.core.fender
import ocyan.core.utils.fixture
import ocyan.core.utils.itertools
import ocyan.core.utils.rudiments
import ocyan.core.utils.theme
import ocyan.core.utils.typeconversion
from configtype.jsonconfig import JsonFinder, setup_search_paths


class FenderConfigTests(unittest.TestCase):
    def tearDown(self):
        sys.meta_path[:] = [
            finder for finder in sys.meta_path if not isinstance(finder, JsonFinder)
        ]
        sys.modules.pop("ocyan.core.json", None)
        importlib.reload(ocyan.core.fender)

    def test_import_without_search_paths_uses_defaults(self):
        sys.meta_path[:] = [
            finder for finder in sys.meta_path if not isinstance(finder, JsonFinder)
        ]
        sys.modules.pop("ocyan.core.json", None)
        module = importlib.reload(ocyan.core.fender)

        self.assertEqual(module.config.settings, {})
        self.assertEqual(module.config.ocyan_plugins, [])
        self.assertTrue(module.config.is_webshop)

    def test_import_with_search_paths_loads_ocyan_json(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            config_path = Path(tmpdir) / "ocyan.json"
            config_path.write_text(
                json.dumps(
                    {
                        "settings": {
                            "django": {
                                "name": "Core test project",
                            },
                            "i18n": {"enabled": True},
                        },
                        "ocyan_plugins": ["barebone", "ocyan.plugin.django"],
                    }
                )
            )

            setup_search_paths(tmpdir)
            sys.modules.pop("ocyan.core.json", None)
            module = importlib.reload(ocyan.core.fender)

            self.assertEqual(module.config.get("django", "name"), "Core test project")
            self.assertTrue(module.config.has_section("i18n"))
            self.assertFalse(module.config.is_webshop)
            self.assertFalse(module.config.i18n_enabled)


def load_tests(loader, tests, ignore):  # pylint: disable=W0613
    tests.addTests(loader.loadTestsFromTestCase(FenderConfigTests))
    tests.addTests(doctest.DocTestSuite(ocyan.core.utils.appextension))
    tests.addTests(doctest.DocTestSuite(ocyan.core.utils.cache))
    tests.addTests(doctest.DocTestSuite(ocyan.core.utils.csp))
    tests.addTests(doctest.DocTestSuite(ocyan.core.utils.currency))
    tests.addTests(doctest.DocTestSuite(ocyan.core.utils.dirtyhacks))
    tests.addTests(doctest.DocTestSuite(ocyan.core.utils.fixture))
    tests.addTests(doctest.DocTestSuite(ocyan.core.utils.itertools))
    tests.addTests(doctest.DocTestSuite(ocyan.core.utils.theme))
    tests.addTests(doctest.DocTestSuite(ocyan.core.utils.typeconversion))
    return tests



__all__ = ["CSPSetting"]


class CSPSetting(object):
    """
    Used to concatenate the CSP settings for the security plugin

    >>> first = CSPSetting()
    >>> klaas = CSPSetting(default_src=["'self'"])
    >>> harrie = first + klaas
    >>> harrie.default_src
    ["'self'"]
    >>> test = harrie + CSPSetting(default_src=["klatsieharrie.nl"], img_src=["testing.nl", "example.com"])
    >>> sorted(test.default_src)
    ["'self'", 'klatsieharrie.nl']
    >>> sorted(test.img_src)
    ['example.com', 'testing.nl']

    >>> test += CSPSetting(style_src=["knaken.nl"])
    >>> sorted(test.style_src)
    ['knaken.nl']
    """

    # pylint: disable=dangerous-default-value
    def __init__(
        self,
        default_src=[],
        script_src=[],
        style_src=[],
        connect_src=[],
        img_src=[],
        font_src=[],
        object_src=[],
        child_src=[],
        form_action=[],
        frame_ancestors=[],
        excluded_url_prefixes=[],
        base_uri=[],
    ):
        self._default_src = set(default_src)
        self._script_src = set(script_src)
        self._style_src = set(style_src)
        self._connect_src = set(connect_src)
        self._img_src = set(img_src)
        self._font_src = set(font_src)
        self._object_src = set(object_src)
        self._child_src = set(child_src)
        self._form_action = set(form_action)
        self._frame_ancestors = set(frame_ancestors)
        self._excluded_url_prefixes = set(excluded_url_prefixes)
        self._base_uri = set(base_uri)

    def __add__(self, other):
        return CSPSetting(
            default_src=self._default_src | other._default_src,
            script_src=self._script_src | other._script_src,
            style_src=self._style_src | other._style_src,
            connect_src=self._connect_src | other._connect_src,
            img_src=self._img_src | other._img_src,
            excluded_url_prefixes=self._excluded_url_prefixes
            | other._excluded_url_prefixes,
            font_src=self._font_src | other._font_src,
            object_src=self._object_src | other._object_src,
            child_src=self._child_src | other._child_src,
            form_action=self._form_action | other._form_action,
            frame_ancestors=self._frame_ancestors | other._frame_ancestors,
            base_uri=self._base_uri | other._base_uri,
        )

    @property
    def default_src(self):
        return list(self._default_src)

    @property
    def script_src(self):
        return list(self._script_src)

    @property
    def style_src(self):
        return list(self._style_src)

    @property
    def connect_src(self):
        return list(self._connect_src)

    @property
    def img_src(self):
        return list(self._img_src)

    @property
    def font_src(self):
        return list(self._font_src)

    @property
    def object_src(self):
        return list(self._object_src)

    @property
    def child_src(self):
        return list(self._child_src)

    @property
    def form_action(self):
        return list(self._form_action)

    @property
    def frame_ancestors(self):
        return list(self._frame_ancestors)

    @property
    def excluded_url_prefixes(self):
        return tuple(self._excluded_url_prefixes)

    @property
    def base_uri(self):
        return list(self._base_uri)



import re
import unicodedata

__all__ = ["module_to_dict", "str_to_identifier"]


def module_to_dict(module):
    """Mimics IMPORT_STAR behavior.

    Allows all symbols from the module not starting with '_'
    """
    return {key: getattr(module, key) for key in dir(module) if not key.startswith("_")}


def str_to_identifier(somestring):
    """Make a valid Python identifier from a string.

    See https://stackoverflow.com/questions/3303312/how-do-i-convert-a-string-to-a-valid-variable-name-in-python

    >>> str_to_identifier('32v2 äßß #Gmw845éh$W b53wi- ')
    '_32v2_a__Gmw845eh_W_b53wi__'
    """
    interpolated_string = (
        unicodedata.normalize("NFKD", somestring)
        .encode("ascii", "ignore")
        .decode("ascii")
    )

    return re.sub(r"\W|^(?=\d)", "_", interpolated_string)



"""Ocyan plugins support

Uses the straight.plugin loader mechanism to load modules from the
ocyan.plugin namespace.
ie the modules ocyan.plugin.wagtail and ocyan.plugin.oscar will be loaded
automatically and installed in Django for you.
"""

from straight.plugin import load

from ocyan.core.sassfunctions import static
from ocyan.core.utils import (
    CSPSetting,
    FixtureAccumulator,
    SegmentedInstalledApps,
    unique_everseen,
)

PLUGINS_NAMESPACE = "ocyan.plugin"


class OcyanPluginConfig:
    """Base class for plugins.
    Used to load plugin configurations in ocyan.main
    """

    def contribute_to_urlpatterns(self, urlpatterns):
        """Mangle urlpatterns to work for this plugin.
        This is a NOOP implementation to allow piping with straight.plugin
        """
        return urlpatterns

    def contribute_to_installed_apps(self, installed_apps: SegmentedInstalledApps):
        """Mangle INSTALLED_APPS to work for this plugin.
        This is a NOOP implementation to allow piping with straight.plugin
        """
        return installed_apps

    def contribute_to_middleware(self, middleware):
        """Mangle MIDDLEWARE to work for this plugin.
        This is a NOOP implementation to allow piping with straight.plugin
        """
        return middleware

    def contribute_to_settings(self, settings):
        """Return the settings for this plugin.
        This is a NOOP implementation to allow piping with straight.plugin
        """
        return settings

    def contribute_to_i18n_fields(self, fields_dict):
        """
        Hook for registering translatable fields for use by the i18n plugin.

        The dictionary looks like this::

            {"app_label.model.field": to_fallback_or_not_to_fallback }

        to_fallback_or_not_to_fallback can be True or False depending if you need empty
        translations to return the base language value instead of None. In most
        cases you want the fallback, but you can disable it if needed.
        """
        return fields_dict

    def contribute_to_fixtures(self, fixtures):
        """
        Hook for adding fixtures to the demo data.

        Usage::

            fixtures.add("app_label", "app_label/things")
            return fixtures

        If you want to override some existing fixture, specify a priority::

            fixtures.add("app_label", "app_label/things", 1)
            return fixtures
        """
        return fixtures

    def get_smoketest_modules(self):
        """Return the test modules from this plugin that should be ran in a
        live site test context.
        """
        return []

    def get_spooler_modules(self, modules):
        "Append spooler modules to the list if they exist"
        return modules

    def contribute_to_sass_functions(self, sass_functions):
        "Add sass functions to be used in your templates to the context"
        return sass_functions

    def contribute_to_csp_settings(self, csp_settings):
        "Add configuration to the CSPSetting object"
        return csp_settings

    def contribute_to_maintenance(self, maintenances):
        """
        Add manintenance jobs

        Maintenance jobs can be added like this::

        def contribute_to_maintenance(self, maintenances):
            priority = 1000
            return maintenances + [
                (priority, "ocyan.core.cool_maintence_function"),
            ]
        """
        return maintenances


def get_plugins(cls=OcyanPluginConfig):
    """Load the plugins."""
    return load(PLUGINS_NAMESPACE, cls).produce()


def get_installed_apps(**initial_apps):
    _installed_apps = SegmentedInstalledApps(**initial_apps)
    plugins = get_plugins()
    return plugins.pipe("contribute_to_installed_apps", _installed_apps)


def get_middleware(*initial_middlewares):
    plugins = get_plugins()
    middleware = plugins.pipe("contribute_to_middleware", list(initial_middlewares))
    return list(unique_everseen(middleware))


def get_fixtures():
    plugins = get_plugins()
    return plugins.pipe("contribute_to_fixtures", FixtureAccumulator())


def add_plugin_settings(settings):
    plugins = get_plugins()
    new_settings = plugins.pipe("contribute_to_settings", settings)
    # because the settings are the local variables of the settings module the
    # result of the settings call should be used to update the settings, because
    # some of the plugin might not mutate the settings dict.
    settings.update(new_settings)


def get_spooler_mdodules():
    plugins = get_plugins()
    return plugins.pipe("get_spooler_modules", [])


def get_spooler_modules():
    return get_spooler_mdodules()


def get_i18n_fields():
    plugins = get_plugins()
    return plugins.pipe("contribute_to_i18n_fields", {})


def get_sass_functions():
    plugins = get_plugins()
    return plugins.pipe("contribute_to_sass_functions", {"static": static})


def get_csp_settings():
    plugins = get_plugins()
    return plugins.pipe("contribute_to_csp_settings", CSPSetting())


def get_maintenances():
    plugins = get_plugins()
    return [m[1] for m in sorted(plugins.pipe("contribute_to_maintenance", []))]



# -*- coding: utf-8 -*-
import collections
import json
from collections.abc import Collection
from itertools import chain

from django.apps.config import AppConfig
from django.core.exceptions import ImproperlyConfigured
from django.core.serializers.json import DjangoJSONEncoder

from .itertools import find_first, unique_everseen

__all__ = ["SegmentedInstalledApps"]


class AppNameTable:
    """
    This class is used to have unified app names. App names in INSTALLED_APPS
    can have the form oscar.apps.shipping.apps.ShippingConfig or just oscar.apps.shipping.
    For purposes of replacement and extension, there should only be 1 way of identifying
    an app, so we normalize the app names in this table.
    """

    def __init__(self, installed_apps=None):
        self._apps = {}
        if installed_apps:
            for installed_app_name in installed_apps:
                self.add(installed_app_name)

    def get_app_config(self, installed_app_name):
        try:
            return AppConfig.create(installed_app_name)
        except (ModuleNotFoundError, ImportError, ImproperlyConfigured, AttributeError):
            return None

    def add(self, installed_app_name):
        app_config = self.get_app_config(installed_app_name)
        if app_config is not None:
            self._apps[installed_app_name] = app_config.name
            self._apps[app_config.name] = app_config.name
        else:
            self._apps[installed_app_name] = installed_app_name

    def get(self, installed_app_name):
        if not installed_app_name:
            return None

        if installed_app_name not in self._apps:
            self.add(installed_app_name)

        return self._apps[installed_app_name]


class ExtendableInstalledAppsItem(list):
    """
    To allow multiple apps to contribute to one and the same oscar app, it
    is needed to setup a registry for this. This class implements the registry.
    >>>
    """

    def __init__(
        self, iterable=(), extends=None, replaced_apps=None, app_name_table=None
    ):
        super(ExtendableInstalledAppsItem, self).__init__(iterable)
        self.extensions = collections.defaultdict(list) if extends is None else extends
        self.replaced_apps = set() if replaced_apps is None else replaced_apps
        self.app_name_table = (
            AppNameTable(replaced_apps) if app_name_table is None else app_name_table
        )

    def normalize_app_name(self, app_name):
        if not app_name:
            return None

        return self.app_name_table.get(app_name)

    def is_section_for_app(self, app_name_):
        """
        >>> INSTALLED_APPS = ExtendableInstalledAppsItem(["oscar.apps.catalogue", "zupke.henkie"])
        >>> INSTALLED_APPS.is_section_for_app("koe")
        False
        >>> INSTALLED_APPS.is_section_for_app("oscar.apps.catalogue")
        True
        >>> INSTALLED_APPS.is_section_for_app("zupke")
        False
        >>> INSTALLED_APPS.is_section_for_app("zupke.henkie")
        True
        >>> INSTALLED_APPS.register_extension("koe", "rarare.vogels")
        >>> INSTALLED_APPS.is_section_for_app("koe")
        True
        >>> INSTALLED_APPS.is_section_for_app("rarare.vogels")
        True
        >>> INSTALLED_APPS.is_section_for_app("rarare.vog")
        False
        >>> INSTALLED_APPS.append("zakhooi.catalogue", replaces="oscar.apps.catalogue")
        >>> INSTALLED_APPS.is_section_for_app("oscar.apps.catalogue")
        True
        >>> INSTALLED_APPS.is_section_for_app("zakhooi.catalogue")
        True
        >>> INSTALLED_APPS.append("keigaaf.catalogue_extension", extends="oscar.apps.catalogue")
        >>> INSTALLED_APPS.is_section_for_app("keigaaf.catalogue_extension")
        True
        >>> INSTALLED_APPS.is_section_for_app("oscar.apps.cat")
        False
        >>> INSTALLED_APPS.register_extension("oscar.apps.cat", "oscar.apps.catalogue")
        >>> INSTALLED_APPS.is_section_for_app("oscar.apps.cat")
        True
        """
        app_name = self.normalize_app_name(app_name_)
        listed_apps = chain(self, chain.from_iterable(self.extensions.values()))
        return (
            app_name in self.replaced_apps
            or app_name in self.extensions
            or any((app_name == app for app in listed_apps))
        )

    def extend(self, installed_app_names):
        """
        Add apps to the list
        """
        for installed_app_name in installed_app_names:
            self.append(installed_app_name)

    def __iadd__(self, installed_app_names):
        """
        Use this to append in place to installed apps.

        >>> INSTALLED_APPS = ExtendableInstalledAppsItem(["oscar.apps.catalogue", "zupke.henkie"])

        We can just add stuff to the INSTALLED_APPS like  this
        >>> INSTALLED_APPS += ["rikje.harkmaken" , "nog.ee.app"]
        >>> INSTALLED_APPS
        ['oscar.apps.catalogue', 'zupke.henkie', 'rikje.harkmaken', 'nog.ee.app']

        If apps are replaced, they are not added again when using +=

        >>> INSTALLED_APPS.append("zakhooi.catalogue", replaces="oscar.apps.catalogue")
        >>> INSTALLED_APPS
        ['zupke.henkie', 'rikje.harkmaken', 'nog.ee.app', 'zakhooi.catalogue']
        >>> INSTALLED_APPS += ["oscar.apps.catalogue", "dongel.type"]
        >>> INSTALLED_APPS
        ['zupke.henkie', 'rikje.harkmaken', 'nog.ee.app', 'zakhooi.catalogue', 'dongel.type']
        """
        self.extend(installed_app_names)
        return self

    def __add__(self, installed_app_names):
        """
        Use this to add to the INSTALLED_APPS

        >>> INSTALLED_APPS = ExtendableInstalledAppsItem(["oscar.apps.catalogue", "zupke.henkie"])

        We can just add stuff to the INSTALLED_APPS like  this
        >>> INSTALLED_APPS + ["rikje.harkmaken" , "nog.ee.app"]
        ['oscar.apps.catalogue', 'zupke.henkie', 'rikje.harkmaken', 'nog.ee.app']

        If apps are replaced, they are not added again when using +=

        >>> INSTALLED_APPS.append("zakhooi.catalogue", replaces="oscar.apps.catalogue")
        >>> INSTALLED_APPS
        ['zupke.henkie', 'zakhooi.catalogue']
        >>> INSTALLED_APPS + ["oscar.apps.catalogue", "dongel.type"]
        ['zupke.henkie', 'zakhooi.catalogue', 'dongel.type']
        """
        new_thing = ExtendableInstalledAppsItem(
            self,
            extends=self.extensions,
            replaced_apps=self.replaced_apps,
            app_name_table=self.app_name_table,
        )
        new_thing.extend(installed_app_names)
        return new_thing

    def register_extension(self, module_name_, extended_app_):
        """
        Register a non replacing app extension, this extension is NOT added
        to INSTALLED_APPS.

        >>> INSTALLED_APPS = ExtendableInstalledAppsItem(["oscar.apps.catalogue", "zupke.henkie"])
        >>> INSTALLED_APPS.register_extension("ocyan.plugin.oscar_catalogue.extras", "oscar.apps.catalogue")
        >>> INSTALLED_APPS
        ['oscar.apps.catalogue', 'zupke.henkie']
        >>> INSTALLED_APPS.extensions_for_app("oscar.apps.catalogue")
        ['ocyan.plugin.oscar_catalogue.extras']

        If you try something stupid it won't work
        >>> INSTALLED_APPS.register_extension("oscar.apps.catalogue", "oscar.apps.catalogue")
        >>> INSTALLED_APPS.extensions_for_app("oscar.apps.catalogue")
        ['ocyan.plugin.oscar_catalogue.extras']

        You can register multiple extensions for the same app
        >>> INSTALLED_APPS.register_extension("gekkehenkie.slimmejongen", "oscar.apps.catalogue")
        >>> INSTALLED_APPS.extensions_for_app("oscar.apps.catalogue")
        ['ocyan.plugin.oscar_catalogue.extras', 'gekkehenkie.slimmejongen']

        You can even register extension for app extensions themselves
        >>> INSTALLED_APPS.register_extension("gekkehenkie.rarejongen",'gekkehenkie.slimmejongen')
        >>> INSTALLED_APPS.extensions_for_app("oscar.apps.catalogue")
        ['gekkehenkie.rarejongen', 'ocyan.plugin.oscar_catalogue.extras', 'gekkehenkie.slimmejongen']
        """
        module_name = self.normalize_app_name(module_name_)
        extended_app = self.normalize_app_name(extended_app_)
        if module_name != extended_app:
            self.extensions[extended_app].append(module_name)

    def append(self, installed_app_name, replaces=None, extends=()):
        """
        Add an app to installed apps. The app might want to replace or extend
        an existing oscar app.

        If your app extension needs to change models, you MUST replace the app.
        This means your app will be put in INSTALLED_APPS instead of the app
        it replaces. You have the responsibility to deal with migrations, urls
        views and admin registrations associated with the app you are replacing.

        If your app extension does NOT do anything with models, you can just
        use extends, your app will be installed next to the replacing app

        :param replaces: The module name of the oscar app the appended app replaces (optional)
        :param extends: List or single instance of oscar app modules names which the appended app extends.
        :type extends: Either a list or a single app module name as a string.

        >>> INSTALLED_APPS = ExtendableInstalledAppsItem(["oscar.apps.catalogue", "zupke.henkie"])
        >>> INSTALLED_APPS
        ['oscar.apps.catalogue', 'zupke.henkie']
        >>> INSTALLED_APPS.append("zakhooi.catalogue", replaces="oscar.apps.catalogue")
        >>> INSTALLED_APPS
        ['zupke.henkie', 'zakhooi.catalogue']
        >>> INSTALLED_APPS.append("keigaaf.catalogue_extension", extends="oscar.apps.catalogue")
        >>> INSTALLED_APPS
        ['zupke.henkie', 'zakhooi.catalogue', 'keigaaf.catalogue_extension']
        >>> INSTALLED_APPS.extensions_for_app("oscar.apps.catalogue")
        ['zakhooi.catalogue', 'keigaaf.catalogue_extension']

        If the order of adding is wrong apps should still chain to the root app

        >>> INSTALLED_APPS = ExtendableInstalledAppsItem(["oscar.apps.catalogue", "zupke.henkie"])
        >>> INSTALLED_APPS.append("keigaaf.catalogue_extension", replaces="deze.app.is.er.nog.niet")
        >>> INSTALLED_APPS
        ['oscar.apps.catalogue', 'zupke.henkie', 'keigaaf.catalogue_extension']
        >>> INSTALLED_APPS.append("deze.app.is.er.nog.niet", replaces="oscar.apps.catalogue")
        >>> INSTALLED_APPS
        ['zupke.henkie', 'keigaaf.catalogue_extension']

        >>> INSTALLED_APPS = ExtendableInstalledAppsItem(["django.contrib.auth"])
        >>> INSTALLED_APPS.append("django.contrib.auth.apps.AuthConfig", replaces="django.contrib.contenttypes.apps.ContentTypesConfig")
        >>> INSTALLED_APPS.extensions_for_app("django.contrib.contenttypes")
        ['django.contrib.auth']

        >>> INSTALLED_APPS = ExtendableInstalledAppsItem(["django.contrib.auth"])
        >>> INSTALLED_APPS.append("django.contrib.auth.apps.AuthConfig", replaces="django.contrib.contenttypes")
        >>> INSTALLED_APPS.extensions_for_app("django.contrib.contenttypes")
        ['django.contrib.auth']

        >>> INSTALLED_APPS = ExtendableInstalledAppsItem(["django.contrib.auth"])
        >>> INSTALLED_APPS.append("django.contrib.auth.apps.AuthConfig", replaces="kloekie.henkie")
        >>> INSTALLED_APPS.extensions_for_app("kloekie.henkie")
        ['django.contrib.auth']
        """
        if isinstance(extends, str):
            extends = [extends]

        if replaces is not None:
            self.replaced_apps.add(replaces)
            # remove apps that are replaced from INSTALLED_APPS
            while replaces in self:
                self.remove(replaces)

            # register the extension
            self.register_extension(installed_app_name, replaces)

        for extended_app in extends:
            # register the extensions
            self.register_extension(installed_app_name, extended_app)

        if installed_app_name not in self.replaced_apps:
            super(ExtendableInstalledAppsItem, self).append(installed_app_name)

    def _extensions_for_app(self, app_name_):
        app_name = self.normalize_app_name(app_name_)
        extensions = self.extensions.get(app_name, [])
        result = []
        for extension in extensions:
            result.extend(self._extensions_for_app(extension))

        return unique_everseen(result + extensions)

    def _history_depth(self, app_name_, target_app_):
        "Determine how many replacements are between app_name and target_app"
        app_name = self.normalize_app_name(app_name_)
        target_app = self.normalize_app_name(target_app_)

        if app_name == target_app:
            return 1  # the target_app was found omg lets count this one!

        # get all the extensions for app_name
        extensions = self.extensions.get(app_name, [])

        # determine miximim distance between extensions and target_app
        current_depth = 0
        for extension in extensions:
            found_depth = self._history_depth(extension, target_app)
            if found_depth > current_depth:
                current_depth = found_depth

        # if the current_depth has changed, increment it, otherwise, just
        # return 0, because the target_app was not found.
        return current_depth + 1 if current_depth else 0

    def extensions_from_app_root(self, app_name):
        """
        Find the app root for ``app_name`` and return the extensions from
        the app root.

        The root of an app is the original application, before it was replaced.

        >>> INSTALLED_APPS = ExtendableInstalledAppsItem(["oscar.apps.catalogue", "zupke.henkie"])
        >>> INSTALLED_APPS.extensions_from_app_root("zupke.henkie")
        ['zupke.henkie']
        >>> INSTALLED_APPS.append("henk.den.berter", replaces="zakhooi.catalogue")
        >>> INSTALLED_APPS.append("zakhooi.catalogue", replaces="oscar.apps.catalogue")
        >>> INSTALLED_APPS.register_extension("ocyan.plugin.oscar_catalogue.extras", "zakhooi.catalogue")
        >>> INSTALLED_APPS.extensions_from_app_root("henk.den.berter")
        ['henk.den.berter', 'ocyan.plugin.oscar_catalogue.extras', 'zakhooi.catalogue', 'oscar.apps.catalogue']
        """
        current_depth = 0
        root = app_name
        for repl in self.replaced_apps:
            depth = self._history_depth(repl, app_name)
            if depth > current_depth:
                current_depth = depth
                root = repl

        return self.extensions_for_app(root) + [root]

    def extensions_for_app(self, app_name):
        """
        Return all the extensions that should be searched for classes for the
        given app_name.

        >>> INSTALLED_APPS = ExtendableInstalledAppsItem(["oscar.apps.catalogue", "zupke.henkie"])
        >>> INSTALLED_APPS.register_extension("ocyan.plugin.oscar_catalogue.extras", "oscar.apps.catalogue")
        >>> INSTALLED_APPS.extensions_for_app("oscar.apps.catalogue")
        ['ocyan.plugin.oscar_catalogue.extras']
        >>> INSTALLED_APPS.append("zakhooi.catalogue", replaces="oscar.apps.catalogue")
        >>> catalogue_extensions = INSTALLED_APPS.extensions_for_app("oscar.apps.catalogue")
        >>> catalogue_extensions
        ['ocyan.plugin.oscar_catalogue.extras', 'zakhooi.catalogue']
        >>> INSTALLED_APPS.append("keigaaf.catalogue_extension", extends="oscar.apps.catalogue")
        >>> INSTALLED_APPS.extensions_for_app("oscar.apps.catalogue")
        ['ocyan.plugin.oscar_catalogue.extras', 'zakhooi.catalogue', 'keigaaf.catalogue_extension']
        >>> INSTALLED_APPS.register_extension("liljon.skeet.skeet", "keigaaf.catalogue_extension")
        >>> INSTALLED_APPS.extensions_for_app("oscar.apps.catalogue")
        ['liljon.skeet.skeet', 'ocyan.plugin.oscar_catalogue.extras', 'zakhooi.catalogue', 'keigaaf.catalogue_extension']
        >>> INSTALLED_APPS.register_extension("third.level.extrensions", "liljon.skeet.skeet")
        >>> INSTALLED_APPS.extensions_for_app("oscar.apps.catalogue")
        ['third.level.extrensions', 'liljon.skeet.skeet', 'ocyan.plugin.oscar_catalogue.extras', 'zakhooi.catalogue', 'keigaaf.catalogue_extension']
        """
        return list(self._extensions_for_app(app_name))

    @property
    def representation(self):
        return {
            "INSTALLED_APPS": self,
            "extensions": {
                app: self.extensions_for_app(app) for app in self.extensions
            },
            "replaced_apps": list(self.replaced_apps),
            "raw": self.extensions,
        }


def oscar_sort_order(oscar_apps):
    """
    Put the dashboard apps at the end.

    >>> apps = [
    ...    "oscar.apps.dashboard.henk",
    ...    "oscar.apps.henk",
    ...    "oscar.apps.klakkie",
    ...    "oscar.apps.dashboard.klakkie",
    ...    "oscar"
    ... ]
    >>> sorted_apps = oscar_sort_order(apps)
    >>> list(sorted_apps)
    ['oscar.apps.henk', 'oscar.apps.klakkie', 'oscar.apps.dashboard.henk', 'oscar.apps.dashboard.klakkie', 'oscar']
    >>> apps = [
    ...    "oscar.apps.dashboard.klakkie",
    ...    "oscar.apps.dashboard.henk",
    ...    "oscar.apps.klakkie",
    ...    "oscar.apps.henk",
    ...    "oscar"
    ... ]
    >>> sorted_apps = oscar_sort_order(apps)
    >>> list(sorted_apps)
    ['oscar.apps.klakkie', 'oscar.apps.henk', 'oscar.apps.dashboard.klakkie', 'oscar.apps.dashboard.henk', 'oscar']
    """
    dashboard_apps = []
    oscar_found = False
    oscar_app_found = False
    for oscar_app_name in oscar_apps:
        if isinstance(oscar_app_name, str) and "dashboard" in oscar_app_name:
            dashboard_apps.append(oscar_app_name)
        elif oscar_app_name == "oscar":
            oscar_found = True
        elif oscar_app_name == "ocyan.plugin.oscar.apps.Shop":
            oscar_app_found = True
        else:
            yield oscar_app_name
    for oscar_app_name in dashboard_apps:
        yield oscar_app_name

    if oscar_found:
        yield "oscar"

    if oscar_app_found:
        yield "ocyan.plugin.oscar.apps.Shop"


class SegmentedInstalledApps(Collection):
    """
    A DataStructure that can be iterated over like a list, but supports
    different named segments: `front`, `django`, `apps`, `oscar`, `wagtail`.

    >>> henk = SegmentedInstalledApps(front=[1,2,3,2], oscar=['ha', 'oscar', 'ha', 2])
    >>> len(henk)
    8
    >>> 1 in henk
    True
    >>> 'ha' in henk
    True
    >>> 'hoe' in henk
    False
    >>> 4 in henk
    False
    >>> list(henk)
    [1, 2, 3, 2, 'ha', 'ha', 2, 'oscar']
    >>> henk.unique_list()
    [1, 2, 3, 'ha', 'oscar']
    >>> henk.ocyan.append("koe")
    >>> list(henk)
    [1, 2, 3, 2, 'koe', 'ha', 'ha', 2, 'oscar']
    >>> henk.ocyan.append("runt")
    >>> list(henk)
    [1, 2, 3, 2, 'koe', 'runt', 'ha', 'ha', 2, 'oscar']
    >>> henk.unique_list()
    [1, 2, 3, 'koe', 'runt', 'ha', 'oscar']
    """

    def __init__(
        self,
        front=(),
        adminoverrides=(),
        django=(),
        ocyan=(),
        apps=(),
        oscar=(),
        wagtail=(),
    ):
        self.front = ExtendableInstalledAppsItem(front)
        self.adminoverrides = ExtendableInstalledAppsItem(adminoverrides)
        self.django = ExtendableInstalledAppsItem(django)
        self.ocyan = ExtendableInstalledAppsItem(ocyan)
        self.apps = ExtendableInstalledAppsItem(apps)
        self.oscar = ExtendableInstalledAppsItem(oscar)
        self.wagtail = ExtendableInstalledAppsItem(wagtail)
        self.segments = (
            self.front,
            self.adminoverrides,
            self.django,
            self.ocyan,
            self.apps,
            self.oscar,
            self.wagtail,
        )

    def unique_list(self):
        return list(unique_everseen(self))

    def section_for_app_extension(self, app_name):
        try:
            return find_first(lambda x: x.is_section_for_app(app_name), self.segments)
        except StopIteration:
            return None

    def __iter__(self):
        "Keep oscar apps sorted"
        return chain(
            self.front,
            self.adminoverrides,
            self.django,
            self.ocyan,
            self.apps,
            oscar_sort_order(self.oscar),
            self.wagtail,
        )

    def __contains__(self, value):
        for segment in self.segments:
            if value in segment:
                return True

        return False

    def __len__(self):
        return sum(map(len, self.segments))

    def __repr__(self):
        return "<SegmentedInstalledApps %s />" % json.dumps(
            {
                "front": self.front.representation,
                "adminoverrides": self.adminoverrides.representation,
                "django": self.django.representation,
                "ocyan": self.ocyan.representation,
                "apps": self.apps.representation,
                "oscar": self.oscar.representation,
                "wagtail": self.wagtail.representation,
            },
            indent=2,
            cls=DjangoJSONEncoder,
        )



from ocyan.core.fender import config

__all__ = ["is_configured_theme"]


def is_configured_theme(theme_name):
    """
    Is 'theme_name' the theme configured in the ocyan.json file.
    """
    assert theme_name
    return config.get("theme", "name", "") == theme_name



from pathlib import Path

from setuptools import find_packages, setup

PACKAGE_CLASSIFIERS = [
    "Framework :: Ocyan",
]

install_requires: list = [
    "Django>=5,<6",
    "straight.plugin>=1.5.1.post1",
    "configtype",
]

extras_require: dict = {
    "test": [
        "wheel",
        "empty_testproject",
        "pylint-django",
        "isort",
        "ruff",
        "vdt.versionplugin.wheel",
        "ocyan.plugin.testing",
        "coverage",
    ]
}

setup(
    name="ocyan.core",
    version="1.2.8",
    url="https://git.mandelblog.com/salt/ocyan.core",
    author="Motolani Olaiya",
    author_email="motolaniolaiya@gmail.com",
    description="A framework for ocyan plugins, all plugins depend on this.",
    long_description=Path("README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.12,<3.13",
    install_requires=install_requires,
    classifiers=PACKAGE_CLASSIFIERS,
    extras_require=extras_require,
)
